#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>

#include "myAllocator.h"


/* how much memory to ask for */
#define DEFAULT_BRKSIZE 0x100000	/* 1M */

//size for displaying to DIV by
#define DIV_BY			1024

/* lowest & highest address in arena (global vars) */
BlockPrefix_t *arenaBegin = (void *)0;
void *arenaEnd = 0;
BlockPrefix_t *nextFitBegin =  (void *)0;


/* create a block, mark it as free */
BlockPrefix_t *makeFreeBlock(void *addr, size_t size) { 
  BlockPrefix_t *p = addr;
  void *limitAddr = addr + size;
  BlockSuffix_t *s = limitAddr - align8(sizeof(BlockSuffix_t));
  p->suffix = s;
  s->prefix = p;
  p->allocated = 0;
  return p;
}

//inits the arena
void initializeArena() {
    if (arenaBegin != 0)	/* only initialize once */
		return;
    arenaBegin = makeFreeBlock(sbrk(DEFAULT_BRKSIZE), DEFAULT_BRKSIZE);
    arenaEnd = ((void *)arenaBegin) + DEFAULT_BRKSIZE;
    nextFitBegin = arenaBegin;
    fprintf(stderr, "Initializing arena from 0x%x to 0x%x, size %dk\n",
			arenaBegin, arenaEnd, (DEFAULT_BRKSIZE/DIV_BY));

}

//clears the arena and re-initializes, by sbrk-ing more space
void clearArena(){
	arenaBegin = (void *)0;
	arenaEnd = 0;
	nextFitBegin =  (void *)0;
	initializeArena();

}

/* useful space within a block */
size_t computeUsableSpace(BlockPrefix_t *p) { 
    void *prefix_end = ((void*)p) + prefixSize;
    return ((void *)(p->suffix)) - (prefix_end);
}

BlockPrefix_t *computeNextPrefixAddr(BlockPrefix_t *p) { 
    return ((void *)(p->suffix)) + suffixSize;
}

BlockSuffix_t *computePrevSuffixAddr(BlockPrefix_t *p) {
    return ((void *)p) - suffixSize;
}

/* return addr of next block (prefix), or 0 if last */
BlockPrefix_t *getNextPrefix(BlockPrefix_t *p) { 
    BlockPrefix_t *np = computeNextPrefixAddr(p);
    if ((void*)np < (void *)arenaEnd)
		return np;
    else
		return (BlockPrefix_t *)0;
}

/* return addr of prev block, or 0 if first */
BlockPrefix_t *getPrevPrefix(BlockPrefix_t *p) { 
    BlockSuffix_t *ps = computePrevSuffixAddr(p);
    if ((void *)ps > (void *)arenaBegin)
		return ps->prefix;
    else
		return (BlockPrefix_t *)0;
}

/* coalesce p with prev, return prev if coalesced, otherwise p */
BlockPrefix_t *coalescePrev(BlockPrefix_t *p) {	
    BlockPrefix_t *prev = getPrevPrefix(p);
    if (p && prev && (!p->allocated) && (!prev->allocated)) {
		makeFreeBlock(prev, ((void *)computeNextPrefixAddr(p)) - (void *)prev);
		return prev;
    }
    return p;
}    

/* coalesce p with prev & next */
void coalesce(BlockPrefix_t *p) {	
    if (p != (void *)0) {
        BlockPrefix_t *next;
		p = coalescePrev(p);
		next = getNextPrefix(p);
		if (next) 
	    	coalescePrev(next);
    }
}

/* split p into two regions A and B, leaving 's' usable space in A and returning B */
BlockPrefix_t *split(BlockPrefix_t *p, size_t s) {	
	//do nothin if in use, or not a block, or s==0, or s>=space(p)
	if (!p || !s || p->allocated || (s >= computeUsableSpace(p)) )
		return p;
	
	//total size of p, including suffix+prefix	
	size_t oldSize = ((void*)p->suffix - (void*)p) + suffixSize;
	//make A, leaving enough space for s and pref and suff
	BlockPrefix_t *A = makeFreeBlock(p, (s+prefixSize+suffixSize)); 
	//make the rest to B
	BlockPrefix_t *B = makeFreeBlock(getNextPrefix(A), oldSize-(s+prefixSize+suffixSize));

	return B;
}

int growingDisabled = 0;	/* true: don't grow arena! (needed for cygwin) */

BlockPrefix_t *growArena(size_t s) { /* this won't work under cygwin since runtime uses brk()!! */
    void *n;
    BlockPrefix_t *p;
    
    if (growingDisabled)
		return (BlockPrefix_t *)0;
    
    s += (prefixSize + suffixSize);
    if (s < DEFAULT_BRKSIZE)
		s = DEFAULT_BRKSIZE;
    
    n = sbrk(s);
    
    if ((n == 0) || (n != arenaEnd)) /* fail if brk moved or failed! */
		return 0;
    
    arenaEnd = n + s;		/* new end */
    p = makeFreeBlock(n, s);	/* create new block */
    p = coalescePrev(p);	/* coalesce with old arena end  */
    
    return p;
}


int pcheck(void *p) {		/* check that pointer is within arena */
    return (p >= (void *)arenaBegin && p < (void *)arenaEnd);
}


void arenaCheck() {		/* consistency check */
    BlockPrefix_t *p = arenaBegin;
    size_t amtFree = 0; size_t amtAllocated = 0;
    int numBlocks = 0;
    
    while (p != 0) {		/* walk through arena */
		fprintf(stderr, "  checking from 0x%x, size=%d, allocated=%d...\n",
			p, computeUsableSpace(p), p->allocated);
		
		assert(pcheck(p));	/* p must remain within arena */
		assert(pcheck(p->suffix)); /* suffix must be within arena */
		assert(p->suffix->prefix == p);	/* suffix should reference prefix */
		
		if (p->allocated) 	/* update allocated & free space */
	    	amtAllocated += computeUsableSpace(p);
		else
	    	amtFree += computeUsableSpace(p);
		
		numBlocks += 1;
		p = computeNextPrefixAddr(p);
		if (p == arenaEnd)
	    	break;
		else
	    	assert(pcheck(p));

    }

    fprintf(stderr, " mcheck: numBlocks=%d, amtAllocated=%db(%dk), amtFree=%db(%dk), arenaSize=%dk\n",
	    numBlocks,
	    amtAllocated, (amtAllocated/DIV_BY),
	    amtFree, (amtFree/DIV_BY),
	    ((((void*)arenaEnd) - ((void*)arenaBegin)) / DIV_BY));

}

/* conversion between blocks & regions (offset of prefixSize */

BlockPrefix_t *regionToPrefix(void *r) {
  if (r)
    return r - prefixSize;
  else
    return 0;
}


void *prefixToRegion(BlockPrefix_t *p) {
  void * vp = p;
  if (p)
    return vp + prefixSize;
  else
    return 0;
}


/* find first block with usable space > s, beginning at arenaBegin each time */
BlockPrefix_t *findFirstFit(size_t s) {	
    BlockPrefix_t *p = arenaBegin;
    while (p) {
		if (!p->allocated && computeUsableSpace(p) >= s)
	    	return p;
		p = getNextPrefix(p);
    }
    return growArena(s);
}

/* find first block with usable space > s, beginning where we last left of */
BlockPrefix_t *findNextFit(size_t s) {	
    BlockPrefix_t *p = nextFitBegin;
    BlockPrefix_t *start = p;
    int repeat = 1;
    
    while (p != start || repeat) {
        repeat = 0;
	
		if (!p->allocated && computeUsableSpace(p) >= s)
	    	return p;
		
		p = getNextPrefix(p);
		if (p == (void*)0)
			p = arenaBegin;
			
		nextFitBegin = p;
    }
    
    return growArena(s);
}

/* find smallest block with usable space >= s */
BlockPrefix_t *findBestFit(size_t s) {	
    BlockPrefix_t *p = arenaBegin;
    BlockPrefix_t *rtnval = (void*)0;
    size_t min_s = DEFAULT_BRKSIZE;
    
    while (p) {
		if (!p->allocated && computeUsableSpace(p) >= s && computeUsableSpace(p) < min_s){
	    	rtnval = p;
	    	min_s = computeUsableSpace(p);
	    }
		p = getNextPrefix(p);
    }
    
    if(rtnval)
    	return rtnval;
    
    return growArena(s);
}


/* this really are equivalent to malloc & free */
void *AllocRegion(size_t s){
	return AllocRegionPtr(s,&findNextFit); 
}


void freeRegion(void *r) {
    if (r != 0) {
		BlockPrefix_t *p = regionToPrefix(r); 		// convert to block
		p->allocated = 0;							// mark as free
		coalesce(p);
    }
}

/* This allocs a region using the supplied find*Fit Function to find the next free block */
void *AllocRegionPtr(size_t s, findFitFunc ffunc) {
  size_t asize = align8(s);
  BlockPrefix_t *p;
  
  if (arenaBegin == 0)		// arena uninitialized?
    initializeArena();
  
  p = (*ffunc)(s);		// find a block using the supplied findFit function
  
  if (p) {			// found a block
    size_t availSize = computeUsableSpace(p);			//the space in block
    if (availSize >= (asize + prefixSize + suffixSize)) // if size is bigger than we need -> split block!
	  	split(p,asize);
    
    p->allocated = 1;		// mark as allocated
    return prefixToRegion(p);	// return converted to *region
  
  }else			// failed
    return (void *)0; 
}


/*
  like realloc(r, newSize), resizeRegion_old will return a new region of size
   newSize containing the old contents of r by:
   1. checking if the present region has sufficient available space to
   satisfy the request (if so, do nothing)
   2. allocating a new region of sufficient size & copying the data
   TODO: if the successor 's' to r's block is free, and there is sufficient space
   in r + s, then just adjust sizes of r & s.
*/
void *resizeRegion_old(void *r, size_t newSize) {
  int oldSize;
  if (r != (void *)0)		/* old region existed */
    oldSize = computeUsableSpace(regionToPrefix(r));
  else
    oldSize = 0;		/* non-existant regions have size 0 */
  if (oldSize >= newSize)	/* old region is big enough */
    return r;
  
  else {			/* allocate new region & copy old data */
    void *n = AllocRegionPtr(newSize,&findNextFit);
	memcpy(n,r,oldSize);
	freeRegion(r);		/* free old region */
	return n;
  }
}

//supplies a standard nextFit for resizing Algo
void *resizeRegion(void *r, size_t newSize) {
	return resizeRegionPtr(r,newSize,&findNextFit);
}

/* resizeRegion that takes a pointer to an alloc func that is used if regions can't be merged
	checks if next region is free and big anough to be merged with current to satisfy req
	if so, it is merged and maybe also split if next region happens to be too big
*/
void *resizeRegionPtr(void *r, size_t newSize, findFitFunc ffunc) {
	int oldSize;
	BlockPrefix_t *r_p = regionToPrefix(r);
	BlockPrefix_t *n_p = getNextPrefix(r_p);
	
	if (r != (void *)0)		// old region existed
		oldSize = computeUsableSpace(r_p);
	else
		oldSize = 0;		// non-existant regions have size 0

	if (oldSize >= newSize)	// old region is big enough
		return r;
	else {			// allocate new region & copy old data
		 //checks if next region is unused and large enough to be merged
		if(!n_p->allocated && (computeUsableSpace(n_p) >= (newSize - oldSize) ) ){

			BlockPrefix_t *p = NULL;
			//if next region + current is larger, and can thus be broken
			if(computeUsableSpace(n_p) >= (newSize - oldSize) )
				//need to split next region on end of reqed size ...
				split(n_p, newSize - oldSize);
			
			p = coalescePrev(n_p);
			return p;
			
		} else {
			void *n = AllocRegionPtr(newSize,ffunc);
			memcpy(n,r,oldSize);
			freeRegion(r);		// free old region
			return n;
		}
	}
}
